#!/usr/bin/perl

# wiki-to-latex.pl - Convert Usemod wiki pages to LaTeX

# Tuomas Toivonen <toivotuo@kasvua.org>
# $Id: wiki-to-latex.pl,v 1.12 2004/05/26 20:35:25 tmtoivon Exp $

use strict;
use IO::File;

# handle command line parameters

if ( $#ARGV != 2 ) {
    die "usage: $0 <wikidb> <docdir> <document>\n";
}

my ($wikidb, $docdir, $document) = @ARGV;
my ($initial) = $document =~ /^(.)/;

# my $wikidir = sprintf "%s/page/%s/%s", $wikidb, $initial, $document;
# my $docdir = sprintf "%s/%s", $docdir, lc $document;

my $wikidir = $wikidb;
my $docdir = $docdir;

# work, work

my @files = &read_directory($wikidir);

foreach my $src (@files) {
    
    my $dst = $src;
    $dst =~ s/^(.+)\.db$/wiki_$1.tex/;
    my $fh = new IO::File;
    $fh->open("> $docdir/$dst")
	or die "Failed to open file '$docdir/$dst' for writing: $!\n";

    my $section = $src;
    $section =~ s/\.db$//;
    $section =~ tr/_/ /;

    $fh->printf("\\section{%s}\n", $section);

    my $text = &fetch_text("$wikidir/$src");
    &parse_text($text, $fh);

    $fh->close;
}

sub parse_text
{
    my ($text, $fh) = @_;

    my $ul_level = 0; # bullet list level
    my $ol_level = 0; # numbered list level
    my $dl_level = 0; # definition list level
    my $tabular = 0; # whether a table is being processed
    my $verbatim = 0; # whether a verbatim block is being processed

    $text =~ s/\r//g; # weed \r from sources

    foreach ( split /\n/, $text ) {	

	next if m|^//|;

        if (not $verbatim) {
            s/_/\\_/g; # quote underscores

	    s/\'{3}(.+?)\'{3}/\{\\bf $1\}/g; # bold text
	    s/\'{2}(.+?)\'{2}/\{\\it $1\}/g; # italic text

	    s//\'/g; # fix apostrophes

        }
        elsif (not /^<\/pre>/) {
	    $fh->print("$_\n");
            next;
        }


	# sections
	if ( /^=\s+(.+?)\s+=/ ) {
	    $fh->print(&do_heading($1, "subsection"));
	}
        
	elsif ( /^==\s+(.+?)\s+==/ ) {
	    $fh->print(&do_heading($1, "subsubsection"));
	}
        
	# paragraphs
	elsif ( /^===\s+(.+?)\s+===/ ) {
	    $fh->print(&do_heading($1, "paragraph"));
	}
        
	# subparagraphs
	elsif ( /^====\s+(.+?)\s+====/ ) {
	    $fh->print(&do_heading($1, "subparagraph"));
	}

	# bullet lists
	elsif ( /^(\*{1,4})\s*(.+)$/ ) {
	    my $level = length $1;
	    if ( $level > $ul_level ) {
                for (1 .. $level - $ul_level) {
                    $fh->print("\\begin{itemize}\n");
                }
	    }
	    elsif ( $level < $ul_level ) {
                for (1 .. $ul_level - $level) {
                    $fh->print("\\end{itemize}\n");
                }
	    }
	    $fh->printf("\\item %s\n", $2);
	    $ul_level = $level;
	}

	# numbered lists
	elsif ( /^(\#{1,4})\s*(.+)$/ ) {
	    my $level = length $1;
	    if ( $level > $ol_level ) {
                for (1 .. $level - $ol_level) {
                    $fh->print("\\begin{enumerate}\n");
                }
	    }
	    elsif ( $level < $ol_level ) {
                for (1 .. $ol_level - $level) {
                    $fh->print("\\end{enumerate}\n");
                }
	    }
	    $fh->printf("\\item %s\n", $2);
	    $ol_level = $level;
	}

	# definition lists
	elsif ( /^(;{1,4})(.+?):(.+)$/ ) {
	    my $level = length $1;
	    if ( $level > $dl_level ) {
                for (1 .. $level - $dl_level) {
                    $fh->print("\\begin{description}\n");
                }
	    }
	    elsif ( $level < $dl_level ) {
                for (1 .. $dl_level - $level) {
                    $fh->print("\\end{description}\n");
                }
	    }
	    $fh->printf("\\item[%s] %s\n", $2, $3);
	    $dl_level = $level;
	}

	# tables
	elsif ( /^\|\|/ ) {
	    my @columns = split /\s*\|\|\s*/;
	    shift @columns; # weed extra element due to leading ||

	    if ( not $tabular ) {
		$fh->printf("\\begin{tabular}{%s}\n", "l" x ($#columns+1));
		$tabular = 1;
	    }
	    $fh->printf("%s \\\\\n", join(" & ", @columns));
	}

        # <pre> -tags (verbatim mode)
        
        elsif ( /^<pre>/ ) {
            $fh->printf("\\begin{verbatim}\n");
            $verbatim = 1;
        }

        elsif ( /^<\/pre>/ ) {
            $fh->printf("\\end{verbatim}\n");
            $verbatim = 0;
        }

	# regular text
	else {
	    while ( $ul_level ) { # standard, close bullet list
		$fh->print("\\end{itemize}\n");
		$ul_level--;
	    }

	    while ( $ol_level ) { # standard, close numbered list
		$fh->print("\\end{enumerate}\n");
		$ol_level--;
	    }

	    while ( $dl_level ) { # standard, close definition list
		$fh->print("\\end{description}\n");
		$dl_level--;
	    }

	    if ( $tabular ) { # close tabulating environment
		$fh->printf("\\end{tabular}\n");
		$tabular = 0;
	    }

	    # handle latex special characters

	    s/#/\\#/g;

	    # output
		
	    $fh->print("$_\n");
	}
    }

    while ( $ul_level ) { # end of document, close bullet list
	$fh->print("\\end{itemize}\n");
	$ul_level--;
    }

    while ( $ol_level ) { # end of document, close numbered list
	$fh->print("\\end{enumerate}\n");
	$ol_level--;
    }

    while ( $dl_level ) { # end of document, close definition list
	$fh->print("\\end{description}\n");
	$dl_level--;
    }
}

sub fetch_text {
    my ($file) = @_;

    my %page = &split_page(&read_file($file));
    my %section = &split_section(%page);
    my %text = &split_text(%section);

    return $text{text};
}

sub do_heading
{
    my ($title, $type) = @_;
 
    if ( $title =~ /^\*/ ) {
	$title = substr $title, 1;
	return sprintf("\\%s*{%s}\n", $type, $title);
    }

    return sprintf("\\%s{%s}\n", $type, $title);
}

# The code below is originally from the Usemod wiki site.
# http://www.usemod.com/cgi-bin/wiki.pl?ExtractingText
# Hacked to eg. work with the 'strict' pragma.

my $FS  = "\xb3";
my $FS1 = $FS . "1";
my $FS2 = $FS . "2";
my $FS3 = $FS . "3";

sub rewrite {
    my ($directory) = @_;
    
    my @files = read_directory($directory);
    foreach my $file (@files) {
	print "Reading $file...\n";
	my %page = &split_page(&read_file("$directory/$file"));
	my %section = &split_section(%page);
	my %text = &split_text(%section);
	$file =~ s/.db$//;
	print "Writing $file...\n";
	&write_file("$directory/$file", $text{text});
    }
    print "Done.\n";
}

sub read_directory {
    my ($dirname) = @_;
    opendir(DIR, $dirname) or die "can't opendir $dirname: $!";
    my @dots = grep { /\.db$/ && -f "$dirname/$_" } readdir(DIR);
    closedir DIR;
    return @dots;
}

sub read_file {
    my ($filename) = @_;
    my ($data);
    my (%page);
    local $/ = undef;   # Read complete files

    open(IN, "<$filename") or die "can't read $filename: $!";
    $data=<IN>;
    close IN;
    return $data;
}

sub split_page {
    my ($data) = @_;
    my (%page);
    my $FS1 = "\xb3" . "1"; # hack, fix
    %page = split(/$FS1/, $data, -1);  # -1 keeps trailing null fields
    return %page;
}

sub split_section {
    my (%page) = @_;
    my (%section);
    my $FS2 = "\xb3" . "2"; # hack, fix
    %section = split(/$FS2/, $page{text_default}, -1);
    return %section;
}

sub split_text {
    my (%section) = @_;
    my (%text);
    my $FS3 = "\xb3" . "3"; # hack, fix
    %text = split(/$FS3/, $section{data}, -1);
    return %text;
}

sub write_file {
    my ($file, $data) = @_;
    open(F, ">" . $file) or die "can't open $file: $!";
    print F $data;
    close(F);
}
